package net.pricefx.common.apibuilder.chart;

import java.math.BigDecimal;

import net.pricefx.common.api.chart.FilterOperator;
import net.pricefx.common.api.chart.LimitPosition;
import net.pricefx.common.api.chart.LimitType;
import net.pricefx.common.api.chart.SeriesType;
import net.pricefx.common.api.chart.SortType;
import net.pricefx.common.api.chart.WhiskerType;
import net.pricefx.common.apibuilder.BacklinkBuilder;
import net.pricefx.common.apibuilder.IdGenerator;
import net.pricefx.common.apibuilder.Invocations;
import net.pricefx.common.apibuilder.MethodInvocation;

public class AbstractChartSeries<C extends AbstractChartBuilder<C>, S extends AbstractChartSeries<C, S>> extends BacklinkBuilder<C> {
	
	protected Filters<S> filters;
	protected Filters<S> productsFilter;
	protected Filters<S> customersFilter;
	protected Filters<S> aggregationsFilter;

	protected AbstractChartSeries(final C backLink, final String instanceId) {
		super(instanceId, "Series", backLink);
	}
	
	IdGenerator getIdGenerator() {
		return idGenerator;
	}
	
	Invocations<MethodInvocation> getInvocations() {
		return invocations;
	}
	
	@SuppressWarnings("unchecked")
	protected S getThis() {
		return (S) this;
	}
	
	// these are common to all charts
	
	public final S setLabel(final String label) {
		addInvocation("setLabel", label);
		return getThis();
	}
	
	public final S setDatamart(final String datamart) {
		addInvocation("setDatamart", datamart);
		return getThis();
	}
	
	public final S setCurrency(final String currency) {
		addInvocation("setCurrency", currency);
		return getThis();
	}
	
	public final Filters<S> getFilters() { // backward compatibility
		return addFilters();
	}
	
	public final Filters<S> addFilters() {
		return addFilters(FilterOperator.AND);
	}
	
	public final Filters<S> addFilters(final FilterOperator operator) {
		final String id = addInvocation("addFilters", operator);
		if (filters == null) {
			filters = new Filters<>(getThis(), id, "Filters");
		}
		return filters;
	}
	
	public final Filters<S> addProductsFilter() {
		final String id = addInvocation("addProductsFilter");
		if (productsFilter == null) {
			productsFilter = new Filters<>(getThis(), id, "ProductsFilter");
		}
		return productsFilter;
	}
	
	public final Filters<S> addCustomersFilter() {
		final String id = addInvocation("addCustomersFilter");
		if (customersFilter == null) {
			customersFilter = new Filters<>(getThis(), id, "CustomersFilter");
		}
		return customersFilter;
	}

	public final Filters<S> addAggregationsFilter() {
		final String id = addInvocation("addAggregationsFilter");
		if (aggregationsFilter == null) {
			aggregationsFilter = new Filters<>(getThis(), id, "AggregationsFilter");
		}
		return aggregationsFilter;
	}

	public final S addDimFilter(final String dimension, final String value) {
		addInvocation("addDimFilter", dimension, value);
		return getThis();
	}
	
	public final Formula<S> setFormula() {
		final String id = addInvocation("setFormula");
		return new Formula<>(getThis(), id, "Formula");
	}
	
	public final S setGeneratedQueryDto(final String generatedQueryDto) {
		addInvocation("setGeneratedQueryDto", generatedQueryDto);
		return getThis();
	}
	
	// Optional methods, idea is to override them as public (and ideally final) and just call super.method()
	
	protected S setHideDataLabels(final Boolean hideDataLabels) {
		addInvocation("setHideDataLabels", hideDataLabels);
		return getThis();
	}
	
	protected LimitToMeasure<S> setLimitTo(final LimitPosition limitPosition, final BigDecimal amount, final LimitType limitType) {
		final String id = addInvocation("setLimitTo", limitPosition, amount, limitType);
		return new LimitToMeasure<>(getThis(), id, "LimitToMeasure");
	}
	
	protected S setType(final SeriesType seriesType) {
		addInvocation("setType", seriesType);
		return getThis();
	}
	
	protected S withSortByAxisX(final SortType sortType) {
		addInvocation("withSortByAxisX", sortType);
		return getThis();
	}

	protected S withSortByAxisY(final SortType sortType) {
		addInvocation("withSortByAxisY", sortType);
		return getThis();
	}

	protected S setComparisonDim(final String comparisonDimension) {
		addInvocation("setComparisonDim", comparisonDimension);
		return getThis();
	}
	
	protected TrendLine<S> withTrendLine() {
		final String id = addInvocation("withTrendLine");
		return new TrendLine<>(getThis(), id, "TrendLine");
	}

	protected S setAggregationBy(final String aggregationBy) {
		addInvocation("setAggregationBy", aggregationBy);
		return getThis();
	}
	
	protected S setBandBy(final String bandBy) {
		addInvocation("setBandBy", bandBy);
		return getThis();
	}
	
	protected S setShowBandByLegend(final Boolean showBandByLegend) {
		addInvocation("setShowBandByLegend", showBandByLegend);
		return getThis();
	}
	
	protected S setHideOther(final Boolean hideOther) {
		addInvocation("setHideOther", hideOther);
		return getThis();
	}
	
	// setAxisX and setAxisY are a bit more tricky because not all charts want the same return type
	
	private String invokeSetAxisX(final String axisX) {
		return addInvocation("setAxisX", axisX);
	}
	
	protected S protectedSetAxisX(final String axisX) {
		invokeSetAxisX(axisX);
		return getThis();
	}
	
	protected AggrFieldAxis<S> protectedSetAggrFieldAxisX(final String axisX) {
		return new AggrFieldAxis<>(getThis(), invokeSetAxisX(axisX), "AxisX");
	}
	
	protected NonAggrFieldAxis<S> protectedSetNonAggrFieldAxisX(final String axisX) {
		return new NonAggrFieldAxis<>(getThis(), invokeSetAxisX(axisX), "AxisX");
	}
	
	private String invokeSetAxisY(final String axisY) {
		return addInvocation("setAxisY", axisY);
	}
	
	protected S protectedSetAxisY(final String axisY) {
		invokeSetAxisY(axisY);
		return getThis();
	}
	
	protected AggrFieldAxis<S> protectedSetAggrFieldAxisY(final String axisY) {
		return new AggrFieldAxis<>(getThis(), invokeSetAxisY(axisY), "AxisY");
	}
	
	protected NonAggrFieldAxis<S> protectedSetNonAggrFieldAxisY(final String axisY) {
		return new NonAggrFieldAxis<>(getThis(), invokeSetAxisY(axisY), "AxisY");
	}
	
	// Methods for chart series with embedded Boxplots
	
	protected S setShowBoxplot(final Boolean showBoxplot) {
		addInvocation("setShowBoxplot", showBoxplot);
		return getThis();
	}
	
	protected NonAggrFieldAxis<S> setBoxplotAxis() {
		String id = addInvocation("setBoxplotAxis");
		return new NonAggrFieldAxis<>(getThis(), id, "BoxplotAxis");
	}
	
	protected S setWhisker(final WhiskerType whiskerType) {
		addInvocation("setWhisker", whiskerType);
		return getThis();
	}
	
	// AdditionalMeasures support
	
	protected String invokeAddAdditionalMeasure() {
		return addInvocation("addAdditionalMeasure");
	}
	
}
